/**********************************************************************
**  Copyright 2003 Craig Graham <craig@data-uncertain.co.uk>
**  Copyright 2003 Steve Essery <tecknobabble@hotmail.com>
**
**  This file is part of the SubApplet Applet.
**
**  The SubApplet Applet is free software; you can redistribute it
**  and/or modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2 of
**  the License, or (at your option) any later version.
**
**  The SubApplet Applet is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with the SubApplet Applet; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**********************************************************************/
#include "subapplet.h"
#include "debug.h"
#include "aboutform.h"
#include "2uparrow.xpm"
#include "2downarrow.xpm"
#include "info.xpm"
#include "librarynames.h"

#include <qpe/global.h>
#include <qpe/qpeapplication.h>
#include <qpe/config.h>
#include <qpe/resource.h>
#ifdef QWS
#include <qpe/qcopenvelope_qws.h>
#endif

#include <qmainwindow.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qtoolbutton.h>
#include <qmessagebox.h>
#include <qdatetime.h>
#include <qtimer.h>
#include <qpopupmenu.h>
#include <qlabel.h>
#include <stdlib.h>
#include <qpainter.h>
#include <stdio.h>
#include <ctype.h>
#include <qdir.h>
#include <qpe/qlibrary.h>
#include <qpe/taskbarappletinterface.h>

#define HEIGHT 20

/**
 *
 * compareAppletPositions(const void *a, const void *b)
 *
 */
static int compareAppletPositions(const void *a, const void *b)
{
    const TaskbarApplet* aa = *(const TaskbarApplet**)a;
    const TaskbarApplet* ab = *(const TaskbarApplet**)b;
    int d = ab->iface->position() - aa->iface->position();
    if ( d ) return d;
    return QString::compare(ab->library->library(),aa->library->library());
}


/**
 *
 * SubApplet::SubApplet()
 *
 */
SubApplet::SubApplet( QWidget *parent )
    : QToolButton( parent ), m_pAbout(NULL), m_pLayout(NULL),
      mButtonDown(false), mTaskBarConfFound(false), mUseDrawButton(false),
      mUseFlatButton(true), mUseDebug(false), mPosition(10), mPopupDelay(1500)
{
    mDebugHelper = DebugHelper("/tmp/subapplet.log");

    readConfig();

    mDebugHelper.logMessage("SubApplet: entered");

    mDebugHelper.logMessage("SubApplet: about to create popup menu");

    m_pMenu = new QPopupMenu(this);
    m_pMenu->setCheckable(true);

    mDebugHelper.logMessage("SubApplet: about to create pixmaps");
    mPixmapClosed = Resource::loadPixmap( "subapplet/2uparrow" );
    mPixmapOpen = Resource::loadPixmap( "subapplet/2downarrow" );
    mPixmapInfo = Resource::loadPixmap( "subapplet/info" );

    if (mPixmapClosed.isNull())
    {
        mDebugHelper.logMessage("SubApplet: failed to find close pixmap, falling back to internal default");
        mPixmapClosed = QPixmap(i2uparrow_xpm);
    }
    if (mPixmapOpen.isNull())
    {
        mDebugHelper.logMessage("SubApplet: failed to find open pixmap, falling back to internal default");
        mPixmapOpen = QPixmap(i2downarrow_xpm);
    }
    if (mPixmapInfo.isNull())
    {
        mDebugHelper.logMessage("SubApplet: failed to info open pixmap, falling back to internal default");
        mPixmapInfo = QPixmap(info_xpm);
    }

    setToggleButton(true);
    setToggleType(QButton::Toggle);
    setOnIconSet(QIconSet(mPixmapOpen, QIconSet::Small));
    setOffIconSet(QIconSet(mPixmapClosed, QIconSet::Small));
    setOn(false);
    setDown(false);
    setAutoRaise(mUseFlatButton);

    mDebugHelper.logMessage("SubApplet: about to create parentless child");
    m_pChild=new QWidget(NULL,
                         NULL,
			 QWidget::WStyle_Customize | QWidget::WStyle_Tool | 
                         QWidget::WStyle_StaysOnTop | QWidget::WGroupLeader);

    mDebugHelper.logMessage("SubApplet: setting child's geometry");
    
    // TODO: hardcoded height....
    m_pChild->setGeometry(0, 0, qApp->desktop()->width(), HEIGHT);

    mDebugHelper.logMessage("SubApplet: hiding child");
    m_pChild->hide();

    mDebugHelper.logMessage("SubApplet: setting NoFocus for child");
    m_pChild->setFocusPolicy(NoFocus);

    mDebugHelper.logMessage("SubApplet: setting child's caption");
    m_pChild->setCaption("SubApplet");

#ifdef QWS
    // Create our custom qcop channel
    mDebugHelper.logMessage("SubApplet: create the QPE/SubApplet channel");
    m_pChannel = new QCopChannel("QPE/SubApplet", this);

    mDebugHelper.logMessage("SubApplet: connecting channel signal");
    connect(m_pChannel, SIGNAL(received(const QCString&, const QByteArray&)),
            this, SLOT(signalReceived(const QCString&, const QByteArray&)));
#endif

    mDebugHelper.logMessage("SubApplet: creating our own internal timer");
    m_pTimer = new QTimer(this);
    connect(m_pTimer, SIGNAL(timeout()), this, SLOT(heldDown()));

    mDebugHelper.logMessage("SubApplet: working out if TaskBar.conf file exists");
    QDir settingsDir(QDir::homeDirPath()+"/Settings");
    mTaskBarConfFound = QFile::exists(settingsDir.filePath("TaskBar.conf"));
    mDebugHelper.logMessage("SubApplet: mTaskBarConfFound=%d", mTaskBarConfFound ? 1 : 0);

    // Try creating our taskbar right now
    loadApplets();

    mDebugHelper.logMessage("SubApplet: exited");
}


/**
 *
 * SubApplet::~SubApplet()
 *
 */
SubApplet::~SubApplet( ) 
{
    mDebugHelper.logMessage("~SubApplet: entered");

    unloadApplets();

    mDebugHelper.logMessage("~SubApplet: about to delete m_pChild");
    if(m_pChild)
        delete m_pChild;

    mDebugHelper.logMessage("~SubApplet: about to delete m_pAbout");
    if(m_pAbout)
        delete m_pAbout;

    mDebugHelper.logMessage("~SubApplet: about to delete m_pMenu");
    if(m_pMenu)
        delete m_pMenu;

    mDebugHelper.logMessage("~SubApplet: about to delete m_pTimer");
    if(m_pTimer)
    {
        if (m_pTimer->isActive()) m_pTimer->stop();
	delete m_pTimer;
    }

    m_pChild = NULL;
    m_pAbout = NULL;
    m_pMenu  = NULL;
    m_pTimer = NULL;

    mDebugHelper.logMessage("~SubApplet: exited");
}


/**
 *
 * SubApplet::readConfig()
 *
 */
void SubApplet::readConfig()
{
    mDebugHelper.logMessage("readConfig: entered");

    // Clear the lists down 
    mLibNames.clear();
    mNiceNames.clear();
    mLibNameDict.clear();
    mNiceNameDict.clear();

    QDir settingsDir(QDir::homeDirPath()+"/Settings");
    QString configFilePath(settingsDir.filePath("SubApplet.conf"));

    Config cfgFile(configFilePath, Config::File);
    cfgFile.setGroup("Settings");

    if (QFile::exists(configFilePath))
    {
        mDebugHelper.logMessage("readConfig: found our config file");

        mLibNames  = cfgFile.readListEntry("FileLibNames", ',');
	mDebugHelper.logMessage("readConfig: read in FileLibNames");
	mNiceNames = cfgFile.readListEntry("DisplayLibNames", ','); 
	mDebugHelper.logMessage("readConfig: read in DisplayLibNames");
	mExcludeApplets = cfgFile.readListEntry("ExcludeApplets", ',');
	mDebugHelper.logMessage("readConfig: read in ExcludeApplets");

	// Check to see if there are any libraries this version of the
	// SubApplet knows about that aren't in the config file, i.e. if we
	// are upgraded then add in the default values for the new applet
	// libraries we recognise to the existing config file
	int i=0;
	int updated=false;
	while (libraryNames[i] != 0)
	{
	    if (!mLibNames.contains(libraryNames[i]))
	    {
	        mLibNames.append(libraryNames[i]);
		mNiceNames.append(friendlyNames[i]);
		if (!updated) updated = true;
	    }
	    i++;
	}

	// Update the config file accordingly
	if (updated)
	{
	    cfgFile.writeEntry("FileLibNames", mLibNames, ',');
	    cfgFile.writeEntry("DisplayLibNames", mNiceNames, ',');
	}

	// Check for debug flag
        mUseDebug = cfgFile.readBoolEntry("Debug", false);
	mDebugHelper.setDebugFlag(mUseDebug);

	// Check for DepressButton flag
	mUseDrawButton = cfgFile.readBoolEntry("DrawButton", false);

	// Check for MakeFlat flag
	mUseFlatButton = cfgFile.readBoolEntry("FlatButton", true);

	// Check for the Position value
	mPosition = cfgFile.readNumEntry("Position", 10);

	// Check for the Popup Delay value
	mPopupDelay = cfgFile.readNumEntry("PopupDelay", 1500);
    }
    else
    {
        // No config file, so construct a default list of shared libraries
	// and human-friendly names
	mDebugHelper.logMessage("readConfig: no config file present, creating default settings");

	int i=0;

	while (libraryNames[i] != 0)
	{
	    mLibNames.append(libraryNames[i]);
	    mNiceNames.append(friendlyNames[i]);
	    i++;
	}

	// Write the lists out to create the file
	cfgFile.writeEntry("FileLibNames", mLibNames, ',');
	cfgFile.writeEntry("DisplayLibNames", mNiceNames, ',');

	// Write out the other settings
	cfgFile.writeEntry("DrawButton", mUseDrawButton);
	cfgFile.writeEntry("FlatButton", mUseFlatButton);
	cfgFile.writeEntry("Position",   mPosition);
	cfgFile.writeEntry("Debug",      mUseDebug);
	cfgFile.writeEntry("PopupDelay", mPopupDelay);
    }

    // Now check for any applet shared libraries we don't know about
    QString path = QPEApplication::qpeDir() + "plugins/applets";
    QDir dir(path, "*.so");
    dir.setFilter(dir.filter()|QDir::Hidden);
    QStringList soLibList = dir.entryList();
    QStringList::Iterator it;

    for (it = soLibList.begin(); it != soLibList.end(); ++it)
    {
	QString libname = *it;

	// Strip of any leading dot
	if (libname.left(1) == ".")
	{
	    libname.remove(0,1);
	}
	// See if the library isn't in the list, but ignore our own library
	if (mLibNames.findIndex(libname) == -1 && libname != "libsubapplet.so")
	{
	    // Its not so add the library into the lists
	    mLibNames.append(libname);

	    // Try to come up with a nice name
	    QString niceName;

	    // Remove any leading "lib"
	    if (libname.left(3) == "lib")
	    {
		niceName = libname.mid(3);
	    }

	    // Remove any trailing ".so"
	    int sep = niceName.find(".so");
	    if (sep > 0)
	    {
		niceName.truncate(sep);
	    }

	    // Remove any trailing "applet"
	    sep = niceName.find("applet");
	    if (sep > 0)
	    {
		niceName.truncate(sep);
	    }

	    // Need to upper case the first character
	    QChar c = niceName.at(0);
	    niceName.remove(0,1);
	    niceName.prepend(c.upper());

	    // Finally!
	    mNiceNames.append(niceName);
	}
    }


    QStringList::Iterator libIt = mLibNames.begin();
    QStringList::Iterator nameIt = mNiceNames.begin();

    if (!mLibNames.isEmpty() && !mNiceNames.isEmpty())
    {
	while (libIt != mLibNames.end() && nameIt != mNiceNames.end())
	{
	   mLibNameDict.insert((*nameIt), new QString(*libIt));
	   mNiceNameDict.insert((*libIt), new QString(*nameIt));
	   libIt++;
	   nameIt++;
	}
    }

    mDebugHelper.logMessage("readConfig: exited");
}


/**
 *
 * SubApplet::showAbout()
 *
 */
void SubApplet::showAbout()
{
    mDebugHelper.logMessage("showAbout: entered");

    if (m_pAbout.isNull())
    {
        mDebugHelper.logMessage("showAbout: about to create new form");
        m_pAbout = new AboutForm(this);
    }

    if (!m_pAbout.isNull())
    {
        if (m_pChild->isVisible())
        {
	    m_pChild->hide();
	    setOn(false);
            setDown(false);
            repaint();
        }
        mDebugHelper.logMessage("showAbout: about to show form");
        m_pAbout->checkGeometry(m_pChild->isVisible());
        m_pAbout->showMaximized();
    }
    mDebugHelper.logMessage("showAbout: exited");
}


/**
 *
 * SubApplet::loadApplets()
 *
 */
void SubApplet::loadApplets()
{
    mDebugHelper.logMessage("loadApplets: entered");

    // Open the plugin/applets directory and get a list of the files
    QString path = QPEApplication::qpeDir() + "plugins/applets";

    // Get the dot-prefixed disabled applets
    QDir dir( path, ".lib*.so"); // load all applets that are disabled...
    dir.setFilter(dir.filter()|QDir::Hidden);

    QStringList dotList = dir.entryList();
    QStringList::Iterator it;

    // Have we found a TaskBar config file?
    if (mTaskBarConfFound)
    {
	// Get the "excluded" applets so we integrate with the way
	// Opie/OpenZaurus disables applets
	Config cfg("TaskBar");
	cfg.setGroup("Applets");
        mDebugHelper.logMessage("loadApplets: reading in Opie/OZ excluded applets");
	QStringList excludeList = cfg.readListEntry("ExcludeApplets", ',');

	if (!excludeList.isEmpty())
	{
	    // Append the Exclude list to the dot-renamed list
	    for (it = excludeList.begin(); it != excludeList.end(); ++it)
	    {
		// Does this applet exist in the list if we prepend it with
		// a dot?  If so we don't want to load it twice.  This may
		// be overly paranoid...
		// When it comes to enabling the applet we can look
		// for the .applet.so and if its not present we know its an
		// Opie/OZ disbaled applet and we can remove it from the 
		// exclude list.
	        QString dotName((*it));
	        if (dotList.find(dotName.prepend(".")) == dotList.end())
		{
		    dotList.append(*it);
		}
	    }
	}
    }
    
    // Add info/about button
    QToolButton *button=new QToolButton(m_pChild);
    button->setGeometry(0,0,m_pChild->height(),m_pChild->height());
    button->setIconSet(mPixmapInfo);
    button->setFixedWidth( m_pChild->height() );
    button->setAutoRaise(true);
    connect(button,SIGNAL(released()),SLOT(showAbout()));
    
    // Create a new layout for the applet
    m_pLayout = new QHBoxLayout( m_pChild, 0, 1 );
    m_pLayout->setAutoAdd(TRUE);
    m_pLayout->setDirection(QBoxLayout::LeftToRight);
    
    // Create the layout for the applets
    QWidget *pTray=new QWidget(m_pChild);
    QHBoxLayout *pTrayLayout;
    pTrayLayout = new QHBoxLayout( pTray, 0, 1 );
    pTrayLayout->setAutoAdd(TRUE);
    pTrayLayout->setDirection(QBoxLayout::LeftToRight);
    
    // Create a list to hold the iformation about the applet's shared libraries
    TaskbarApplet **applets = new TaskbarApplet* [dotList.count()];
    int nApplets=0;
    bool hadError=false;
    QStringList faulty;

    // Now iterate of the shared library file list
    for ( it = dotList.begin(); it != dotList.end(); ++it ) 
    {
	// Don't want to move ourselves off the main taskbar
	// We should never see this condition!
        if(*it!="libsubapplet.so" && *it!=".libsubapplet.so")
        {
	    // Is the applet on our ExcludeApplets list
	    QString noDotName = *it;
	    noDotName.remove(0,1);

	    if (mExcludeApplets.findIndex(noDotName) == -1)
	    {
		// Load the shared library as its not on the exclude list
		QString fullName = path + "/" + *it;
		mDebugHelper.logMessage("loadApplets: found \"%s\"", (const char*)fullName);

		QLibrary *lib = new QLibrary( fullName );
		TaskbarAppletInterface *iface = 0;

		// Query the plugin interface
		if ( lib->queryInterface( IID_TaskbarApplet, (QUnknownInterface**)&iface ) == QS_OK ) 
		{
		    // Create the struct to hold the information about this applet
		    applets[nApplets] = new TaskbarApplet;
		    applets[nApplets]->library = lib;
		    applets[nApplets]->iface = iface;
		    nApplets++;
		}
		else
		{
		    // We don't check for this....hmmm
		    hadError=true;

		    // Same as Taskbar settings uses

		    // Chop off the leading "lib"
		    QString name = (*it).mid(3);

		    // If there is a trailing .so chop it off
		    int sep = name.find( ".so" );
		    if ( sep > 0 )
			name.truncate( sep );

		    // 
		    sep = name.find( "applet" );
		    if ( sep == (int)name.length() - 6 )
			name.truncate( sep );
		    name[0] = name[0].upper();
		    faulty += name; 
		    delete lib;
		}
	    }
	    else
	    {
	        mDebugHelper.logMessage("loadApplets: applet %s is on the exclude list, ignoring", noDotName.latin1());
	    }
        }
    }

    printf("SubApplet: found %d applets\n",nApplets);
    
    // Sort the applets, based on their position() method.
    qsort(applets,nApplets,sizeof(applets[0]),compareAppletPositions);
    
    // Insert the applets in sorted order into the child panel
    while(nApplets--)
    {
        // Create instance of applet as child of our m_pChild sub panel
        applets[nApplets]->applet = applets[nApplets]->iface->applet( pTray );
        m_appletList.append(applets[nApplets]);
        QString lang = getenv( "LANG" );
        QTranslator * trans = new QTranslator(qApp);
        QString type = (*it).left( (*it).find(".") );
        QString tfn = QPEApplication::qpeDir()+"/i18n/"+lang+"/"+type+".qm";
        if ( trans->load( tfn ))
            qApp->installTranslator( trans );
        else
            delete trans;
    }
    pTrayLayout->addStretch(10);

    
    delete [] applets;
    mDebugHelper.logMessage("loadApplets: exited");
}


/**
 *
 * SubApplet::unloadApplets(bool deleteLayout)
 *
 */
void SubApplet::unloadApplets()
{
    mDebugHelper.logMessage("unloadApplets: entered");

#ifndef SINGLE_APP
    QValueList<TaskbarApplet*>::Iterator mit;
    for ( mit = m_appletList.begin(); mit != m_appletList.end(); ++mit ) {
        (*mit)->iface->release();
        (*mit)->library->unload();
        delete (*mit)->library;
        delete (*mit);
    }
#endif
    m_appletList.clear();
    if ( m_pLayout )
        delete m_pLayout;

    mDebugHelper.logMessage("unloadApplets: exited");
}


/**
 *
 * SubApplet::mousePressEvent(QMouseEvent *)
 *
 */
void SubApplet::mousePressEvent( QMouseEvent* )
{
    mDebugHelper.logMessage("mousePressEvent: entered");

    // Using our own internal timer to fix problems with the mechanism (we
    // could have a succession of pending signals which could result in the
    // menu being popped up if the user happens to click the button at the 
    // right time
    mButtonDown = true;
    m_pTimer->start(mPopupDelay, true);

    mDebugHelper.logMessage("mousePressEvent: exited");
}

/**
 *
 * SubApplet::mouseReleaseEvent(QMouseEvent *)
 *
 */
void SubApplet::mouseReleaseEvent( QMouseEvent * )
{
    mDebugHelper.logMessage("mouseReleaseEvent: entered");

    mButtonDown = false;

    if(!m_pLayout)
    {
        loadApplets();
    }
    else
    {
	// Fix: changes applets make while on the secondary task bar while it
	// is hidden aren't picked up.  Force the layout to sort that out.
	QValueList<TaskbarApplet*>::Iterator mit;
	for ( mit = m_appletList.begin(); mit != m_appletList.end(); ++mit )
            (*mit)->applet->updateGeometry();
    }

    if (m_pChild->isVisible())
    {
	m_pChild->hide();
	setOn(false);
        setDown(false);

        // TODO: Ajusting about dialog's geometry if it is visible
        if (m_pAbout)
        {
	    if (m_pAbout->isVisible())
	    {
		m_pAbout->setGeometry(m_pAbout->x(), m_pAbout->y(),
				      m_pAbout->width(), m_pAbout->height()+HEIGHT);
		m_pAbout->updateGeometry();
	    }
        }
    }
    else
    {
	m_pChild->resize( qApp->desktop()->width(), m_pChild->height() );
	m_pChild->move( 0, mapToGlobal( QPoint() ).y() - m_pChild->height() );

        // TODO: May need to adjust the about dialog's geometry if it is visible
        if (m_pAbout)
        {
	    if (m_pAbout->isVisible())
	    {
		m_pAbout->setGeometry(m_pAbout->x(), m_pAbout->y(),
				      m_pAbout->width(), m_pAbout->height()-HEIGHT);
		m_pAbout->updateGeometry();
	    }
        }

	m_pChild->show();
	setOn(true);
	setDown(true);
    }
    repaint();

    mDebugHelper.logMessage("mouseReleaseEvent: exited");
}

/**
 *
 * SubApplet::heldDown()
 *
 */
void SubApplet::heldDown()
{
    mDebugHelper.logMessage("heldDown: entered");

    // Reload our config as this will add ensure that the library/name lists
    // are up to date and also reflect any applets that have been added
    readConfig();

    if (mButtonDown && !m_pMenu->isVisible())
    {
        mDebugHelper.logMessage("heldDown: popup menu");

	if(m_pChild->isVisible())
	{
	    m_pChild->hide();
	    update();
	}
	m_pMenu->clear();

	QStringList ozExcludeList;
	if (mTaskBarConfFound)
	{
	    // Get the "excluded" applets so we integrate with the way
	    // Opie/OpenZaurus disables applets
            mDebugHelper.logMessage("heldDown: fetching OZ exclude list");
	    Config cfg("TaskBar");
	    cfg.setGroup("Applets");
	    ozExcludeList = cfg.readListEntry("ExcludeApplets", ',');
	}

        // Need to scan down our library lists
	QStringList::Iterator libIt  = mLibNames.begin();
        QStringList::Iterator nameIt = mNiceNames.begin();

	// Applet Dir
	QDir appletDir(QPEApplication::qpeDir()+"plugins/applets");

	if (!mLibNames.isEmpty() && !mNiceNames.isEmpty())
	{
	    while (libIt != mLibNames.end() && nameIt != mNiceNames.end())
	    {
		QString libName = *libIt;
		QString dotName = libName;
		dotName.prepend(".");
		int  id = -1;

                mDebugHelper.logMessage("heldDown: checking for %s", (*libIt).latin1());

		// Check for the prescence of the the shared library
		if (QFile::exists(appletDir.filePath(libName)))
		{
		    // Library present, no leading dot but need to see if it an
		    // oz Excluded applet
		    if (ozExcludeList.findIndex(libName) == -1)
		    {
			// Not on oz list
                        mDebugHelper.logMessage("heldDown: found, not on oz exclde list");
		        id = m_pMenu->insertItem(*nameIt);
		        m_pMenu->setItemChecked(id, true);
		    }
		    else
		    {
			// On oz list
                        mDebugHelper.logMessage("heldDown: found, on oz exclude list");
		        id = m_pMenu->insertItem(*nameIt);
		        m_pMenu->setItemChecked(id, false);
		    }
		}
		else if (QFile::exists(appletDir.filePath(dotName)))
		{
		    // Library present, leading dot
		    id = m_pMenu->insertItem(*nameIt);
		    m_pMenu->setItemChecked(id, false);
		}

	        libIt++;
		nameIt++;
	    }
	}
	connect(m_pMenu, SIGNAL(activated(int)),
	        this, SLOT(menuItemClicked(int)));

	QPoint curPos = mapToGlobal(QPoint(0,0));
	QSize sh = m_pMenu->sizeHint();
	m_pMenu->popup( curPos-QPoint((sh.width()-width())/2, sh.height()));
    }
}


/**
 *
 * SubApplet::menuItemClicked(int)
 *
 */
void SubApplet::menuItemClicked(int id)
{
    mDebugHelper.logMessage("menuItemClicked: entered");
    QString dirPath;
    QString oldFile;
    QString newFile;

    // Work out the path to $QPEDIR/plugins/applets
    dirPath = QPEApplication::qpeDir();
    dirPath.append("plugins/applets/");

    // Prepend the filenames with the directory paths
    oldFile = dirPath;
    newFile = dirPath;

    // Get the text for the selected entry (shared library filename)
    QString niceName = m_pMenu->text(id);
    QString *libName = mLibNameDict.find(niceName);
    QString fileName = *libName;

    if (m_pMenu->isItemChecked(id))
    {
        // If the menu item is checked we are disabling the applet
        // by adding a leading . to the filename
	// NB  There's no point in adding the shared library to the Opie/OZ
	//     ExcludedApplets list as we have added a . to the library name
	//     and the OZ Launcher won't see the libraries and therefore
	//     doesn't need to exclude them from its TaskBar
	mDebugHelper.logMessage("menuItemClicked: disabling %s", fileName.latin1());
        oldFile.append(fileName);
        newFile.append(".");
        newFile.append(fileName);

	// rename the file!
	rename(oldFile, newFile);
    }
    else
    {
        // If the menu item is unchecked we are enabling the applet
        // by removing the leading . we added to the filename

	// Now need to also check to see if the file is present without 
	// the leading dot, i.e. its an Opie/OZ disabled applet in which
	// case no rename is required (but we need to modify the TaskBar
	// ExcludedApplets list
	mDebugHelper.logMessage("menuItemClicked: enabling %s", fileName.latin1());
        oldFile.append(".");
        oldFile.append(fileName);
        newFile.append(fileName);

	if (QFile::exists(oldFile))
	{
	    // rename the file!
	    mDebugHelper.logMessage("menuItemClicked: renaming %s to %s", oldFile.latin1(), newFile.latin1());
	    rename(oldFile, newFile);
	}
	else
	{
	    // Paranoia
	    if (mTaskBarConfFound)
	    {
		// Load the ExcludedApplets list and remove this one
		mDebugHelper.logMessage("menuItemClicked: removing %s from TaskBar.conf ExcludedApplets list", fileName.latin1());
		Config cfg("TaskBar");
		cfg.setGroup("Applets");
		QStringList excludeList = cfg.readListEntry("ExcludeApplets", ',');
		QStringList newExcludeList;
		QStringList::Iterator it;
		for (it = excludeList.begin(); it != excludeList.end(); ++it)
		{
		    if ( (*it) != fileName )
		    {
			newExcludeList += (*it);
		    }
		}
		cfg.writeEntry("ExcludeApplets", newExcludeList, ',');
	    }
	}
    }

    // Now tell Qtopia to refresh the applets, but not immediately as this
    // seems to make the process a lot more stable.
    // The result of this is that we will be destroyed and recreated
    // i.e. unloadApplets() is called
    QTimer::singleShot(250, this, SLOT(fireReload()));

    mDebugHelper.logMessage("menuItemClicked: exited");
}


/**
 *
 * SubApplet::fireReload(QPaintEvent *)
 *
 */
void SubApplet::fireReload()
{
    mDebugHelper.logMessage("fireReload: entered");

    QCopChannel::send("QPE/TaskBar", "reloadApplets()");

    mDebugHelper.logMessage("fireReload: exited");
}

/**
 *
 * SubApplet::paintEvent(QPaintEvent *)
 *
 */
void SubApplet::paintEvent( QPaintEvent* )
{
    mDebugHelper.logMessage("paintEvent: entered");
    QPainter p(this);

    if (mUseDrawButton)
    {
        mDebugHelper.logMessage("paintEvent: using drawButton()");
        drawButton(&p);
    }
    else
    {
        mDebugHelper.logMessage("paintEvent: not using drawButton()");
	int x=(width()-mPixmapOpen.width())/2;
	int y=(height()-mPixmapOpen.height())/2;
	if(m_pChild->isVisible())
	    p.drawPixmap(x,y,mPixmapOpen);  
	else
	    p.drawPixmap(x,y,mPixmapClosed);  
    }
    mDebugHelper.logMessage("paintEvent: exited");
}


/**
 *
 * SubApplet::
 *
 */
bool SubApplet::useDrawButton()
{
    return mUseDrawButton;
}


/**
 *
 * SubApplet::
 *
 */
bool SubApplet::useFlatButton()
{
    return mUseFlatButton;
}


/**
 *
 * SubApplet::
 *
 */
bool SubApplet::useDebug()
{
    return mUseDebug;
}


/**
 *
 * SubApplet::
 *
 */
int SubApplet::position()
{
    return mPosition;
}


/**
 *
 * SubApplet::
 *
 */
int SubApplet::delay()
{
    return mPopupDelay;
}


/**
 *
 * SubApplet::
 *
 */
void SubApplet::setUseDrawButton(bool flag)
{
    mUseDrawButton = flag;
}


/**
 *
 * SubApplet::
 *
 */
void SubApplet::setUseFlatButton(bool flag)
{
    mUseFlatButton = flag;
}


/**
 *
 * SubApplet::
 *
 */
void SubApplet::setUseDebug(bool flag)
{
    mUseDebug = flag;
    mDebugHelper.setDebugFlag(flag);
}


/**
 *
 * SubApplet::
 *
 */
void SubApplet::setPosition(int pos)
{
    mPosition = pos;
}


/**
 *
 * SubApplet::
 *
 */
void SubApplet::setDelay(int delay)
{
    mPopupDelay = delay;
}


/**
 *
 * SubApplet::
 *
 */
StringDict SubApplet::libNameDict()
{
    return mLibNameDict;
}


/**
 *
 * SubApplet::
 *
 */
StringDict SubApplet::niceNameDict()
{
    return mNiceNameDict;
}


/**
 *
 * SubApplet::
 *
 */
QStringList SubApplet::libNames()
{
    return mLibNames;
}


/**
 *
 * SubApplet::
 *
 */
QStringList SubApplet::niceNames()
{
    return mNiceNames;
}


/**
 *
 * SubApplet::
 *
 */
QStringList SubApplet::excludeApplets()
{
    return mExcludeApplets;
}


#ifdef QWS
/**
 *
 * SubApplet::signalReceived(const QCString& msg, const QByteArray& data)
 *
 */
void SubApplet::signalReceived(const QCString& msg, const QByteArray& data)
{
    mDebugHelper.logMessage("signalReceived: entered");
    QDataStream stream(data, IO_ReadOnly);
    if (msg == "debug(QString)")
    {
        QString flag;
        stream >> flag;
        flag = flag.lower();
        mDebugHelper.logMessage("signalReceived: debug(%s)", flag.latin1());
	setUseDebug((flag == "true"));

    }
    if (msg == "readConfig()")
    {
        mDebugHelper.logMessage("signalReceived: readConfig()");
	readConfig();
    }
    mDebugHelper.logMessage("signalReceived: exited");
}
#endif
